/* wtapfile-format.c - Simple cross-platform wiretap fileformat!
 *
 *
 * Wiretap Library - functions to read/write native wtapfile format
 *
 * Copyright (c)2000 by Axis Communications AB, LUND, SWEDEN
 * Comments/Bugreports should be sent to johan.jorgensen@axis.com
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <sys/time.h>
#include <netinet/in.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "wtap.h"
#include "wtap-int.h"
#include "file_wrappers.h"
#include "buffer.h"
#include "wtap_file-format.h"


typedef struct {
  int wtap_magic;            /* Magic for this fileformat */
  char major_version;        /* Major file version */
  char minor_version;        /* Minor file version */
  unsigned long last_packet; /* Offset in bytes to last packet written */
  int file_type;             /* Type of file e.g. WTAP_FILE_WTAP */
  int snaplen;               /* snaplen */
  int file_encap;            /* Encapsulation */
} wtapfile_hdr_t;


/* ************************************************************************* */
/*                            Forward Declarations                           */
/* ************************************************************************* */
static gboolean wtapfile_dump(wtap_dumper *wdh,
			      const struct wtap_pkthdr *phdr,
			      const  union wtap_pseudo_header *ps_hdr,
			      const u_char *pd, int *err);

static gboolean wtapfile_read(wtap *wth,int *err,int *data_offset);
static gboolean wtapfile_close(wtap_dumper *wdh,int *err);



/* ************************************************************************* */
/* FUNCTION: wtapfile_can_write_encap(int filetype, int encap)               */
/*                                                                           */
/* DESCRIPTION: Called by higher level file management layer in order to     */
/*              determine if the wtap-filetype supports the combination      */
/*              filetype/packetencapsulation. wtap-filetype will support all */
/*              encapsulationtypes an no translations are made to the data   */
/*              written.                                                     */
/*                                                                           */
/* RETURN VALUE: Should return 0 if the filetype/encap type is supported     */
/* ************************************************************************* */
int wtapfile_can_write_encap(int filetype,int encap)
{
  return (filetype == WTAP_FILE_WTAP) ? 0:1;
}



/* ************************************************************************* */
/* FUNCTION: wtapfile_dump_open(wtap_dumper *wdh,int *err)                   */
/*                                                                           */
/* DESCRIPTION: Called by higher level file-management layer in order to     */
/*              open the file. wtapfile_dump_open should set fields in the   */
/*              wdh as a way to provide functions for read/write/close       */
/*              operations.                                                  */
/*                                                                           */
/* RETURN VALUE: Should return TRUE on success and FALSE on error with err   */
/*               set accordingly                                             */
/* ************************************************************************* */
gboolean wtapfile_dump_open(wtap_dumper *wdh,int *err) 
{
  guint32 magic;
  wtapfile_hdr_t file_hdr;
  wtapfile_hdr_t *new_hdr;
  int nwritten;
  magic = WTAP_FILE_MAGIC;
  
  /* Create and store a copy of the fileheader for later use (close) */
  new_hdr = g_malloc(sizeof(wtapfile_hdr_t));

  if (new_hdr == NULL) {
	*err = WTAP_ERR_SHORT_WRITE;
	return FALSE;
  }

  file_hdr.wtap_magic = htonl(magic);
  file_hdr.major_version = WTAPFILE_MAJOR_VERSION;
  file_hdr.minor_version = WTAPFILE_MINOR_VERSION;

  file_hdr.last_packet = 0;                    /* We will be accessing this 
												* field frequently so we do 
												* _not_ convert it */

  file_hdr.file_type = htonl(wdh->file_type); /* Host independent format */
  file_hdr.snaplen = htonl(wdh->snaplen);
  file_hdr.file_encap = htonl(wdh->encap);



  wdh->subtype_write = wtapfile_dump;
  wdh->subtype_close = wtapfile_close;
  
  nwritten = fwrite(&file_hdr,1,sizeof(wtapfile_hdr_t),wdh->fh);
  
  if (nwritten != sizeof(wtapfile_hdr_t)) {
    *err = (nwritten < 0) ? errno:WTAP_ERR_SHORT_WRITE;
    return FALSE;
  }

  /* Copy the fileheader to the stored copy */
  memcpy(new_hdr,&file_hdr,sizeof(wtapfile_hdr_t));
  
  wdh->dump.opaque = (void *) new_hdr;

  return TRUE;
}
  
  

/* ************************************************************************* */
/* FUNCTION: wtapfile_open(wtap *wth, int *err)                              */
/*                                                                           */
/* DESCRIPTION: Called by file-managment layer in wiretap as a result of     */
/*              a call to wtap_open_offline. This function will read and     */
/*              verify the fileheader.                                       */
/*                                                                           */
/* RETURN VALUE: 1 if the file is found to be a "wtap file". Zero otherwise  */
/* ************************************************************************* */
int wtapfile_open(wtap *wth,int *err)
{
  guint32 magic;
  int nread;
  wtapfile_hdr_t file_header;
  wtapfile_hdr_t *this_header;

  file_seek(wth->fh,0,SEEK_SET);
  
  this_header = g_malloc(sizeof(wtapfile_hdr_t));
  wth->capture.generic = this_header;

  if (this_header == NULL) {
	errno = WTAP_ERR_CANT_READ;
	return -1;
  }
  errno = WTAP_ERR_CANT_READ;
  nread = file_read(&file_header,1,sizeof(wtapfile_hdr_t),wth->fh);
  
  
  if (nread != sizeof(wtapfile_hdr_t)) {
    *err = file_error(wth->fh);
    if (*err) return -1;
    return 0;
  }
  
  magic = ntohl(file_header.wtap_magic);

  if ((magic != WTAP_FILE_MAGIC) ||
      (file_header.major_version > WTAPFILE_MAJOR_VERSION)) {
    *err = WTAP_ERR_UNSUPPORTED;
    return -1;
  }
  
  /* Set fields in wtap structure */
  wth->data_offset = sizeof(wtapfile_hdr_t);
  wth->file_type = WTAP_FILE_WTAP;

  wth->snapshot_length = ntohl(file_header.snaplen);
  wth->file_type = ntohl(file_header.file_type);
  wth->file_encap = ntohl(file_header.file_encap);
  
  wth->subtype_read = wtapfile_read;
  wth->subtype_seek_read = wtap_def_seek_read;
  wth->subtype_close = NULL;

  /* Update fields in newly allocated file header */
  this_header->wtap_magic = WTAP_FILE_MAGIC;
  this_header->major_version = WTAPFILE_MAJOR_VERSION;
  this_header->minor_version = WTAPFILE_MINOR_VERSION;
  
  this_header->last_packet = ntohl(file_header.last_packet);
  this_header->file_type = ntohl(file_header.file_type);
  this_header->file_encap = ntohl(file_header.file_encap);
  this_header->snaplen = ntohl(file_header.snaplen);
								  
  return 1;
}



/* ************************************************************************* */
/* FUNCTION: gboolean wtapfile_dump(wtap_dumper *wdh,                        */
/*                                  const struct wtap_pkthdr *phdr,          */
/*			            const  union wtap_pseudo_header *ps_hdr, */
/*			            const u_char *pd, int *err)              */
/*                                                                           */
/* DESCRIPTION: Backend for wtap_dump. Used when writing a file of type      */
/*              wtapfile.                                                    */
/*                                                                           */
/* RETURN VALUE: TRUE if the packet was successfully written. FALSE if not   */
/*               (reason code in *err)                                       */
/* ************************************************************************* */
static gboolean wtapfile_dump(wtap_dumper *wdh,
			      const struct wtap_pkthdr *phdr,
			      const  union wtap_pseudo_header *ps_hdr,
			      const u_char *pd, int *err)
{
  struct wtap_pkthdr phdr_copy;
  int  nwritten; 
  wtapfile_hdr_t *the_header;
 
  /* Get the header from the current dumper */
  the_header = (wtapfile_hdr_t *) wdh->dump.opaque;
  
  /* Convert wtap_pkthdr to host-independent version */
  phdr_copy.ts.tv_sec = htonl(phdr->ts.tv_sec);
  phdr_copy.ts.tv_usec = htonl(phdr->ts.tv_usec);
  phdr_copy.caplen = htonl(phdr->caplen);
  phdr_copy.len = htonl(phdr->len);
  phdr_copy.pkt_encap = htonl(phdr->pkt_encap);
  
  nwritten = fwrite((const voidp)&phdr_copy,1,
					sizeof(struct wtap_pkthdr),wdh->fh);
  if (nwritten != sizeof(struct wtap_pkthdr)) {
	*err = ( nwritten < 0) ? errno: WTAP_ERR_SHORT_WRITE;
	return FALSE;
  }

    
  nwritten = fwrite((const voidp) pd,1,phdr->caplen,wdh->fh);
  if (nwritten != phdr->caplen) {
	*err = ( nwritten < 0) ? errno: WTAP_ERR_SHORT_WRITE;
	return FALSE;
  }
  
  /* Store the length of the last packet saved */
  the_header->last_packet  = phdr->caplen;
  return TRUE;
}



/* ************************************************************************* */
/* FUNCTION: gboolean wtapfile_read(wtap *wth,int *err,int *data_offset)     */
/*                                                                           */
/* DESCRIPTION: Read the next ethereal-information header and the associated */
/*              packet from the file                                         */
/*                                                                           */
/* RETURN VALUE: Returns TRUE on success FALSE on error (error code in err)  */
/* ************************************************************************* */
static gboolean wtapfile_read(wtap *wth,int *err,int *data_offset)
{
  struct wtap_pkthdr phdr_copy;
  wtapfile_hdr_t *the_header;
  int nread;
  int packet_size;
  
  the_header = wth->capture.generic;
  
  /* Check if we are trying to access a packet beyond the end of the file */
  if (the_header->last_packet <= wth->data_offset) { 
	*err = 0;
	return FALSE;
  }

  nread = file_read(&phdr_copy,1,sizeof (struct wtap_pkthdr),wth->fh);

  if (nread != sizeof(struct wtap_pkthdr)) {
    *err = (errno < 0) ? errno:WTAP_ERR_SHORT_READ;
    return FALSE;
  }

  /* Check if packets too long */
  packet_size = ntohl(phdr_copy.caplen);

  if (packet_size>WTAP_MAX_PACKET_SIZE) {
    g_message("Wiretap: File has %u-byte packet. Maximum packet length is %u",
	      ntohl(phdr_copy.caplen),WTAP_MAX_PACKET_SIZE);
    *err = WTAP_ERR_BAD_RECORD;
    return FALSE;
  }

  wth->data_offset += nread;

  /* Convert and store wtap_pkthdr in wth->phdr */
  wth->phdr.ts.tv_sec = ntohl(phdr_copy.ts.tv_sec);
  wth->phdr.ts.tv_usec = ntohl(phdr_copy.ts.tv_usec);
  wth->phdr.caplen = ntohl (phdr_copy.caplen);
  wth->phdr.len = ntohl(phdr_copy.len);
  wth->phdr.pkt_encap = ntohl(phdr_copy.pkt_encap);
 
  /* Read packet from file and store in buffer */
  buffer_assure_space(wth->frame_buffer,packet_size);
  *data_offset = wth->data_offset;
  nread = file_read(buffer_start_ptr(wth->frame_buffer),1, packet_size,
		    wth->fh);

  if (nread != packet_size) {
    *err = (errno < 0) ? errno: WTAP_ERR_SHORT_READ;
    return FALSE;
  }

  wth->data_offset += nread;
  return TRUE;
}


/* ************************************************************************* */
/* FUNCTION: void wtapfile_close(wtap *wth)                                  */
/*                                                                           */
/* DESCRIPTION: Close the dump file and free allocated memory                */
/*                                                                           */
/* RETURN VALUE: none (void)                                                 */
/* ************************************************************************* */
static gboolean wtapfile_close(wtap_dumper *wdh,int *err)
{
  int nbytes;
  wtapfile_hdr_t *this_header;
  unsigned long file_len;

  file_len = ftell(wdh->fh);
  /* Get the file-header */
  this_header = (wtapfile_hdr_t *) wdh->dump.opaque;
  
  /* Use current value of last_packet to calculate last-packet-offset */
  this_header->last_packet = file_len - this_header->last_packet;
  
  
  /* Update fields in the fileheader */
  this_header->last_packet = ntohl(this_header->last_packet);

  /* Write the new header to the file */
  nbytes = fseek(wdh->fh,0,SEEK_SET);
  nbytes = fwrite(this_header,1,sizeof(wtapfile_hdr_t),wdh->fh);

  if (nbytes != sizeof(wtapfile_hdr_t)) {
	*err = WTAP_ERR_SHORT_WRITE;
	return FALSE;
  }
  
  /* Make sure we can truncate file in the future (should it be necessary) */
  fseek(wdh->fh,0,SEEK_END);

  g_free(this_header);
  wdh->dump.opaque = NULL;

  return TRUE;
}
